
from nyto import net
import numpy as np

def _linear(data):
	return data

def _relu(data):
	return np.maximum(0, data)

def _lulu(data):
	'''UC_a1ZYZ8ZTXpjg9xUY9sj8w'''
	return np.log(np.maximum(1,data+1))

def _gaussian(data):
	return np.exp(-data**2)

def _sigmoid(data):
	return 1 / (1 + np.exp(-data))

def _tanh(data):
	return np.tanh(data)

def _col_nor(data):
	return (data-np.mean(data, axis=0))/np.std(data, axis=0)

def _row_nor(data):
	return (
		data-np.mean(data, axis=1)[:,np.newaxis]
	)/np.std(data, axis=1)[:,np.newaxis]

def _softmax(data):
	exp_x=np.exp(data-np.max(data, axis=-1, keepdims=True))
	return exp_x/np.sum(exp_x, axis=-1, keepdims=True)

def _concatenate(*args, axis=1):
	return np.concatenate(args, axis=axis)

def _tile(data, size_tuple):
	return np.tile(data, size_tuple)

def _accuracy(pre_np, target_np):
    return np.mean(pre_np.argmax(axis=1)==target_np.argmax(axis=1))

def _MSE(pre_np, target_np):
	return np.mean(np.square(pre_np-target_np))

def _RMSE(pre_np, target_np):
	return np.sqrt(_MSE(pre_np, target_np))

def _MAE(pre_np, target_np):
	return np.mean(np.abs(pre_np-target_np))

def _MAPE(pre_np, target_np):
	return np.mean(np.abs((pre_np-target_np)/target_np))

def _cross_entropy(pre_np, target_np, epsilon=1e-12):
	pre_np = np.clip(pre_np, epsilon, 1.-epsilon)
	N = pre_np.shape[0]
	ce = - np.sum(target_np*np.log(pre_np)) / N
	return ce

def _binary_cross_entropy(pre_np, target_np, epsilon=1e-12):
	pre_np = np.clip(pre_np, epsilon, 1.-epsilon)
	ce = - np.mean(target_np*np.log(pre_np)+(1-target_np)*np.log(1-pre_np))
	return ce

def _pad(data_np, kernel_shape, mod='valid', constant=0):
	'''
	圖片周圍補值(0)
	[param]
	(np)data_np:          4d_np[data_idx][img_idx][row_idx][col_idx]
	(tuple)kernel_shape:  exp:(3,2)
	(str)mode:            'valid' or 'full'
	(float/int)constant:  填充值
	[return]
	(np)4d_np
	'''
	if mod=='full':
		row_size,col_size=kernel_shape
		row_pad=(row_size-1,row_size-1)
		col_pad=(col_size-1,col_size-1)
		return np.pad(
			data_np,
			((0,0), (0,0), row_pad, col_pad),
			"constant",
			constant_values=(constant, constant)
		)

	if mod=='valid':
		row_size,col_size=kernel_shape
		row_pad_size=int(row_size/2)
		col_pad_size=int(col_size/2)
		row_pad=(row_pad_size,row_pad_size)
		col_pad=(col_pad_size,col_pad_size)
		return np.pad(
			data_np,
			((0,0), (0,0), row_pad, col_pad),
			"constant", 
			constant_values=(constant, constant)
		)

	return data_np

def _fix_pad(data_np, kernel_shape, strides):
	'''
	當移動窗口無法剛好分割時，可以用相鄰的值擴充右邊跟下邊
	[param]
	(np)data_np:          4d_np[data_idx][img_idx][row_idx][col_idx]
	(tuple)kernel_shape:  exp:(3,2)
	(int)strides:         步伐大小(窗口間隔)
	[return]
	(np)4d_np:            [data_idx][img_idx][row_idx][col_idx]
	'''

	(data_shape,img_shape,row_shape,col_shape)=data_np.shape
	(row_split,col_split)=kernel_shape

	row_excess=(row_shape-row_split)%strides
	row_fix=0 if row_excess==0 else strides-row_excess

	col_excess=(col_shape-col_split)%strides
	col_fix=0 if col_excess==0 else strides-col_excess

	return np.pad(
		data_np,
		((0,0), (0,0), (0, row_fix),(0,col_fix)),
		'edge'
	)

def _strided(data_np, kernel_shape, strides=1):
	'''
	將圖片分割成許多小窗口
	[param]
	(np)data_np:          4d_np[data_idx][img_idx][row_idx][col_idx]
	(tuple)kernel_shape:  exp:(3,2)
	(int)strides:         步伐大小(窗口間隔)
	[return]
	(np)6d_np: [data_shape, img_shape, row_stepn, col_stepn, row_split, col_split]
	'''
	data_btyes=data_np.strides[-1]
	(data_shape,img_shape,row_shape,col_shape)=data_np.shape

	(row_split,col_split)=kernel_shape
	col_stepn=((col_shape-col_split)//strides)+1
	row_stepn=((row_shape-row_split)//strides)+1

	return np.lib.stride_tricks.as_strided(
		data_np,
		shape=(data_shape, img_shape, row_stepn, col_stepn, row_split, col_split),
		strides=(
			img_shape*row_shape*col_shape*data_btyes,
			row_shape*col_shape*data_btyes,
			strides*col_shape*data_btyes,
			strides*data_btyes,
			col_shape*data_btyes,
			data_btyes
		)
	)

def _strided_3d(data_np, kernel_shape, strides=1):
	'''
	將圖片分割成許多小窗口
	[param]
	(np)data_np:          4d_np[data_shape][img_shape][row_shape][col_shape]
	(tuple)kernel_shape:  exp:(3,2)
	(int)strides:         步伐大小(窗口間隔)
	[return]
	(np)6d_np: [data_shape, row_stepn, col_stepn, img_shape, row_split, col_split]
	'''
	data_btyes=data_np.strides[-1]
	(data_shape,img_shape,row_shape,col_shape)=data_np.shape

	(row_split,col_split)=kernel_shape
	col_stepn=((col_shape-col_split)//strides)+1
	row_stepn=((row_shape-row_split)//strides)+1

	return np.lib.stride_tricks.as_strided(
		data_np,
		shape=(
			data_shape, 
			row_stepn, col_stepn,
			img_shape,
			row_split, col_split
		),
		strides=(
			img_shape*row_shape*col_shape*data_btyes,
			strides*col_shape*data_btyes,
			strides*data_btyes,
			row_shape*col_shape*data_btyes,
			col_shape*data_btyes,
			data_btyes
		)
	)

def _conv_compute(cat_np, filter_np):
	'''
	計算窗口與kernl的乘積
	[param]
	(np)batch_np:   4d_np[data_shape, cat_shape, row_split, col_split]
	(np)kernals_np: 3d_np[filter_shape, row_split, col_split]

	[return]
	(np)3d_np:      [data_shape, filter_shape, cat_shape]
	'''
	return np.einsum('dckij,fkij -> dfc', cat_np, filter_np)

def _convolution(data_np, filter_np, mod='valid', strides=1):
	'''
	[param]
	(np)data_np:     4d_np[data_shape][img_shape][row_shape][col_shape]
	(np)filter_np:   4d_np[filter_shape][img_shape][row_split][col_split]
	(str)mod:        'full' or 'valid' or 'same'
	(int)strides:    窗口移動間隔
	[return]
	(np)new_data_np: 4d_np[data_shape][conv_shape][row_stepn][col_stepn]
	'''
	(data_shape, img_shape, row_shape, col_shape)=data_np.shape
	(filter_shape, img_shape, row_split, col_split)=filter_np.shape

	# 周圍補0 #
	pad_np=data_np if mod=='same' else _pad(
		data_np=data_np,
		kernel_shape=(row_split,col_split),
		mod=mod,
		constant=0
	)

	# 切割出窗口 #
	strided_np=_strided_3d(
		data_np=pad_np,
		kernel_shape=(row_split,col_split),
		strides=strides
	)

	(_, row_stepn, col_stepn, _, _, _)=strided_np.shape

	# 合併窗口的排列方式: 2d排列 -> 1d排列 #
	cat_np=strided_np.reshape(
		data_shape, -1, img_shape, row_split, col_split,
	)

	# 計算窗口與filter的乘積 #
	conv_np=_conv_compute(cat_np, filter_np)

	# 將卷集完map排列由1d還原回2d #
	return conv_np.reshape(
		data_shape, filter_shape, row_stepn, col_stepn
	)

def _max_pooling(data_np, kernel_shape, strides=1):
	'''
	[param]
		(np)data_np:         4d_np[data_shape][img_shape][row_shape][col_shape]
		(tuple)kernel_shape: exp=(2,3)
		(int)strides:        窗口移動間隔
	[return]
		(np)4d_np:           [data_shape][img_shape][row_stepn][col_stepn]
	'''
	fix_np=_fix_pad(
		data_np=data_np,
		kernel_shape=kernel_shape,
		strides=strides
	)

	strided=_strided(
		data_np=fix_np,
		kernel_shape=kernel_shape,
		strides=strides
	)
	return strided.max(axis=(4,5))    

def _average_pooling(data_np, kernel_shape, strides=1):
	'''
	[param]
		(np)data_np:         4d_np[data_shape][img_shape][row_shape][col_shape]
		(tuple)kernel_shape: exp=(2,3)
		(int)strides:        窗口移動間隔
	[return]
		(np)4d_np:           [data_shape][img_shape][row_stepn][col_stepn]
	'''
	fix_np=_fix_pad(
		data_np=data_np,
		kernel_shape=kernel_shape,
		strides=strides
	)

	strided=_strided(
		data_np=fix_np,
		kernel_shape=kernel_shape,
		strides=strides
	)
	return strided.mean(axis=(4,5))

def _global_average_pooling(data_np):
	'''
	[param]
	(np)data_np: 4d_np[data_shape][img_shape][row_shape][col_shape]
	[return]
	(np)2d_np:   [data_shape][feature_shape]
	'''
	return data_np.mean(axis=(2,3))

def _global_max_pooling(data_np):
	'''
	[param]
	(np)data_np: 4d_np[data_shape][img_shape][row_shape][col_shape]
	[return]
	(np)2d_np:   [data_shape][feature_shape]
	'''
	return data_np.max(axis=(2,3))

def _flattening(data_np):
	'''
	[param]
	(np)data_np: 4d_np[data_shape][img_shape][row_shape][col_shape]
	[return]
	(np)2d_np:   [data_shape][feature_shape]
	'''

	if type(data_np)==list: data_np=np.array(data_np)

	data_shape=data_np.shape[0]
	return data_np.reshape(
		data_shape, int(data_np.size/data_shape)
	)

def add_func_to_data_ref(net_ref, node_id, func):
	if not node_id in net_ref.unit:
		net_ref.data[node_id]=func

	if not node_id in net_ref.node_id_set:
		net_ref.node[node_id]=net.data_node(node_id)

class add_func_node(net.add_func_to_net):
	def __init__(self, node_id, func):
		self.node_id=node_id
		self.func=func

	def __call__(self, node_if):
		net_ref=node_if.net_ref
		add_func_to_data_ref(net_ref, self.node_id, self.func)
		return net_ref[self.node_id](node_if)

class linear(add_func_node):
	def __init__(self): super().__init__(
		node_id='_linear', func=_linear
	)

class relu(add_func_node):
	def __init__(self): super().__init__(
		node_id='_relu', func=_relu
	)

class lulu(add_func_node):
	def __init__(self): super().__init__(
		node_id='_lulu', func=_lulu
	)

class gaussian(add_func_node):
	def __init__(self): super().__init__(
		node_id='_gaussian', func=_gaussian
	)

class sigmoid(add_func_node):
	def __init__(self): super().__init__(
		node_id='_sigmoid', func=_sigmoid
	)

class tanh(add_func_node):
	def __init__(self): super().__init__(
		node_id='_tanh', func=_tanh
	)

class col_nor(add_func_node):
	def __init__(self): super().__init__(
		node_id='_col_nor', func=_col_nor
	)

class row_nor(add_func_node):
	def __init__(self): super().__init__(
		node_id='_raw_nor', func=_row_nor
	)

class softmax(add_func_node):
	def __init__(self): super().__init__(
		node_id='_softmax', func=_softmax
	)

class global_average_pooling(add_func_node):
	def __init__(self): super().__init__(
		node_id='_global_average_pooling',
		func=_global_average_pooling
	)

class global_max_pooling(add_func_node):
	def __init__(self): super().__init__(
		node_id='_global_max_pooling',
		func=_global_max_pooling
	)

class flattening(add_func_node):
	def __init__(self): super().__init__(
		node_id='_flattening',
		func=_flattening
	)

class max_pooling(net.add_func_to_net):
	'''
	node.data_np >> max_pooling(node.kernel_shape, node.strides) >> 
	'''
	def __init__(self, kernel_shape_node_if, strides=1):
		self.kernel_shape_node_if=kernel_shape_node_if
		self.strides_node_if=strides

	def __call__(self, data_np_node_if):
		net_ref=data_np_node_if.net_ref
		func_node_id='_max_pooling'
		add_func_to_data_ref(
			net_ref=net_ref,
			node_id=func_node_id,
			func=_max_pooling
		)
		return net_ref[func_node_id](
			data_np=data_np_node_if,
			kernel_shape=self.kernel_shape_node_if,
			strides=self.strides_node_if
		)

class average_pooling(net.add_func_to_net):
	'''
	node.data_np >> average_pooling(node.kernel_shape, node.strides) >> 
	'''
	def __init__(self, kernel_shape_node_if, strides=1):
		self.kernel_shape_node_if=kernel_shape_node_if
		self.strides_node_if=strides

	def __call__(self, data_np_node_if):
		net_ref=data_np_node_if.net_ref
		func_node_id='_average_pooling'
		add_func_to_data_ref(
			net_ref=net_ref,
			node_id=func_node_id,
			func=_average_pooling
		)
		return net_ref[func_node_id](
			data_np=data_np_node_if,
			kernel_shape=self.kernel_shape_node_if,
			strides=self.strides_node_if
		)

def concatenate(*node_ifs, axis=1):
	net_ref=node_ifs[0].net_ref
	(node_id,func)=('_concatenate',_concatenate)

	add_func_to_data_ref(net_ref, node_id, func)
	return net_ref[node_id](*node_ifs, axis=axis)

def tile(node_if, size_tuple):
	net_ref=node_if.net_ref
	(node_id,func)=('_tile',_tile)

	add_func_to_data_ref(net_ref, node_id, func)
	return net_ref[node_id](node_if, size_tuple)

def MSE(pre_if, target_if):
	net_ref=pre_if.net_ref
	(node_id,func)=('_MSE',_MSE)

	add_func_to_data_ref(net_ref, node_id, func)
	return net_ref[node_id](pre_if, target_if)

def RMSE(pre_if, target_if):
	net_ref=pre_if.net_ref
	(node_id,func)=('_RMSE',_RMSE)

	add_func_to_data_ref(net_ref, node_id, func)
	return net_ref[node_id](pre_if, target_if)

def MAE(pre_if, target_if):
	net_ref=pre_if.net_ref
	(node_id,func)=('_MAE',_MAE)

	add_func_to_data_ref(net_ref, node_id, func)
	return net_ref[node_id](pre_if, target_if)

def MAPE(pre_if, target_if):
	net_ref=pre_if.net_ref
	(node_id,func)=('_MAPE',_MAPE)

	add_func_to_data_ref(net_ref, node_id, func)
	return net_ref[node_id](pre_if, target_if)

def cross_entropy(pre_if, target_if):
	net_ref=pre_if.net_ref
	(node_id,func)=('_cross_entropy',_cross_entropy)

	add_func_to_data_ref(net_ref, node_id, func)
	return net_ref[node_id](pre_if, target_if)

def binary_cross_entropy(pre_if, target_if):
	net_ref=pre_if.net_ref
	(node_id,func)=('_binary_cross_entropy',_binary_cross_entropy)

	add_func_to_data_ref(net_ref, node_id, func)
	return net_ref[node_id](pre_if, target_if)

def accuracy(pre_if, target_if):
	net_ref=pre_if.net_ref
	(node_id,func)=('_accuracy',_accuracy)
	
	add_func_to_data_ref(net_ref, node_id, func)
	return net_ref[node_id](pre_if, target_if)




from nyto import particle
import numpy as np
from nyto import unit_function as uf

class layer(particle.particle):
	def __repr__(self):
		class_name=self.__class__.__name__
		if self.is_fix:
			return f"{class_name}({self.shape}, Fix={self.is_fix})"
		return f"{class_name}({self.shape})"

class variable_layer(layer):
	'''
	[mod]
	(np)values: tag(variable)
	'''
	def __getitem__(self, key):
		return self.values[key]
	
	@property
	def shape(self):
		return self.values.shape
	
	@property
	def values(self):
		return self.mod['values']

	@values.setter
	def values(self, new_var):
		self.mod['values']=new_var

class nn_layer(layer):
	'''
	[mod]
	(np)weights: tag(weights, dropout)
	(np)bias:    tag(bias)
	'''

	def __call__(self, input_data):
		if type(input_data)==np.ndarray:
			return_np=input_data.dot(self.weights)+self.bias
			return return_np

		if type(input_data)==variable_layer:
			return self(input_data.values)

		raise ValueError("input data type is not correct!")
	
	@property
	def shape(self):
		return self.weights.shape

	@property
	def weights(self):
		return self.mod['weights']

	@weights.setter
	def weights(self, new_weights):
		self.mod['weights'] = new_weights

	@property
	def bias(self):
		return self.mod['bias']

	@bias.setter
	def bias(self, new_bias):
		self.mod['bias'] = new_bias

class lstm_layer(layer):
    '''
    [mod]
        (variable_layer)com_in:       tag(weights, dropout)
        (variable_layer)com_mem:      tag(weights, dropout)
        (variable_layer)com_out:      tag(weights, dropout)
        (variable_layer)com_bias:     tag(bias)

        (variable_layer)ingate_in:    tag(weights, dropout)
        (variable_layer)ingate_mem:   tag(weights, dropout)
        (variable_layer)ingate_out:   tag(weights, dropout)
        (variable_layer)ingate_bias:  tag(bias)

        (variable_layer)outgate_in:   tag(weights, dropout)
        (variable_layer)outgate_mem:  tag(weights, dropout)
        (variable_layer)outgate_out:  tag(weights, dropout)
        (variable_layer)outgate_bias: tag(bias)

        (variable_layer)forgate_in:   tag(weights, dropout)
        (variable_layer)forgate_mem:  tag(weights, dropout)
        (variable_layer)forgate_out:  tag(weights, dropout)
        (variable_layer)forgate_bias: tag(bias)

        (variable_layer)init_mem:     tag(init_state)
        (variable_layer)init_out:     tag(init_state)
    [var]
        (def)com_func:                _tanh
        (def)ingate_func:             _sigmoid
        (def)outgate_func:            _sigmoid
        (def)forgate_func:            _sigmoid
        (def)final_func:              _tanh
    '''

    def _neural_compute(self, in_v, mem_v, out_v, in_w, mem_w, out_w, bias):
        return in_v.dot(in_w.values)+mem_v.dot(mem_w.values)+out_v.dot(out_w.values)+bias.values

    def run(self, in_np, mem_np, out_np):
        com_ret=self.com_func(self._neural_compute(
            in_np, mem_np, out_np,
            self.com_in, self.com_mem, self.com_out, self.com_bias
        ))

        ingate_ret=self.ingate_func(self._neural_compute(
            in_np, mem_np, out_np,
            self.ingate_in, self.ingate_mem, self.ingate_out, self.ingate_bias
        ))

        outgate_ret=self.outgate_func(self._neural_compute(
            in_np, mem_np, out_np,
            self.outgate_in, self.outgate_mem, self.outgate_out, self.outgate_bias
        ))

        forgate_ret=self.forgate_func(self._neural_compute(
            in_np, mem_np, out_np,
            self.forgate_in, self.forgate_mem, self.forgate_out, self.forgate_bias
        ))

        new_mem=com_ret*ingate_ret+mem_np*forgate_ret
        final_ret=self.final_func(new_mem)*outgate_ret

        return final_ret, new_mem

    def run_series(self, input_data):
        return_data_list=[]
        last_output=self.init_out.values
        memory_np=self.init_mem.values
        for this_times_data_np in input_data:
            this_times_data_np=this_times_data_np[np.newaxis,:]
            last_output, memory_np=self.run(
                this_times_data_np, memory_np, last_output
            )
            return_data_list.append(last_output.squeeze(0))

        return np.array(return_data_list)

    def __call__(self, batch_data):
        return [self.run_series(data) for data in batch_data]

    @property
    def shape(self):
        return self.com_in.shape

    @property
    def com_in(self): return self.mod['com_in']
    @property
    def com_mem(self): return self.mod['com_mem']
    @property
    def com_out(self): return self.mod['com_out']
    @property
    def com_bias(self): return self.mod['com_bias']

    @com_in.setter
    def com_in(self,x): self.mod['com_in']=x
    @com_mem.setter
    def com_mem(self,x): self.mod['com_mem']=x
    @com_out.setter
    def com_out(self,x): self.mod['com_out']=x
    @com_bias.setter
    def com_bias(self,x): self.mod['com_bias']=x
    
    @property
    def ingate_in(self): return self.mod['ingate_in']
    @property
    def ingate_mem(self): return self.mod['ingate_mem']
    @property
    def ingate_out(self): return self.mod['ingate_out']
    @property
    def ingate_bias(self): return self.mod['ingate_bias']

    @ingate_in.setter
    def ingate_in(self,x): self.mod['ingate_in']=x
    @ingate_mem.setter
    def ingate_mem(self,x): self.mod['ingate_mem']=x
    @ingate_out.setter
    def ingate_out(self,x): self.mod['ingate_out']=x
    @ingate_bias.setter
    def ingate_bias(self,x): self.mod['ingate_bias']=x
    
    @property
    def outgate_in(self): return self.mod['outgate_in']
    @property
    def outgate_mem(self): return self.mod['outgate_mem']
    @property
    def outgate_out(self): return self.mod['outgate_out']
    @property
    def outgate_bias(self): return self.mod['outgate_bias']

    @outgate_in.setter
    def outgate_in(self,x): self.mod['outgate_in']=x
    @outgate_mem.setter
    def outgate_mem(self,x): self.mod['outgate_mem']=x
    @outgate_out.setter
    def outgate_out(self,x): self.mod['outgate_out']=x
    @outgate_bias.setter
    def outgate_bias(self,x): self.mod['outgate_bias']=x
    
    @property
    def forgate_in(self): return self.mod['forgate_in']
    @property
    def forgate_mem(self): return self.mod['forgate_mem']
    @property
    def forgate_out(self): return self.mod['forgate_out']
    @property
    def forgate_bias(self): return self.mod['forgate_bias']

    @forgate_in.setter
    def forgate_in(self,x): self.mod['forgate_in']=x
    @forgate_mem.setter
    def forgate_mem(self,x): self.mod['forgate_mem']=x
    @forgate_out.setter
    def forgate_out(self,x): self.mod['forgate_out']=x
    @forgate_bias.setter
    def forgate_bias(self,x): self.mod['forgate_bias']=x
    
    @property
    def init_mem(self): return self.mod['init_mem']
    @property
    def init_out(self): return self.mod['init_out']

    @init_mem.setter
    def init_mem(self,x): self.mod['init_mem']=x
    @init_out.setter
    def init_out(self,x): self.mod['init_out']=x
    
    @property
    def com_func(self): return self.mod['com_func']
    @property
    def ingate_func(self): return self.mod['ingate_func']
    @property
    def outgate_func(self): return self.mod['outgate_func']
    @property
    def forgate_func(self): return self.mod['forgate_func']
    @property
    def final_func(self): return self.mod['final_func']

    @com_func.setter
    def com_func(self,x): self.mod['com_func']=x
    @ingate_func.setter
    def ingate_func(self,x): self.mod['ingate_func']=x
    @outgate_func.setter
    def outgate_func(self,x): self.mod['outgate_func']=x
    @forgate_func.setter
    def forgate_func(self,x): self.mod['forgate_func']=x
    @final_func.setter
    def final_func(self,x): self.mod['final_func']=x

class conv_layer(layer):
	'''
	[mod]
		(np)filter_np: tag(weights, convolution)
		(np)filter_np: 4d_np[filter_shape, img_shape, row_split, col_split]
	[var]
		(def)conv_func: _convolution
		(str)pad_mod:   'full' or 'valid' or 'same'
		(int)strides:   窗口移動間隔
	'''

	def __call__(self, data_4d_np):
		return self.conv_func(
			data_np=data_4d_np,
			filter_np=self.filter_np,
			mod=self.pad_mod,
			strides=self.strides
		)
	
	@property
	def shape(self):
		(filter_shape, img_shape, row_split, col_split)=self.filter_np.shape
		return (img_shape, filter_shape), (row_split, col_split)
	
	@property
	def filter_np(self): return self.mod['filter_np']
	@filter_np.setter
	def filter_np(self, x): self.mod['filter_np']=x

	@property
	def conv_func(self): return self.var_dict['_convolution']
	@conv_func.setter
	def conv_func(self, x): self.var_dict['_convolution']=x

	@property
	def pad_mod(self): return self.var_dict['pad_mod']
	@pad_mod.setter
	def pad_mod(self, x): self.var_dict['pad_mod']=x

	@property
	def strides(self): return self.var_dict['strides']
	@strides.setter
	def strides(self, x): self.var_dict['strides']=x

def new_variable_layer(
	structure, init_values=0, random_size=None, dropout=False, fix=False,
	tags={'variable'}
):
	new_variable = variable_layer()
	if random_size is None:
		new_variable.values=np.zeros(structure)+init_values
		if dropout: new_variable.tag['dropout']={'values'}
		for tag in tags: new_variable.tag[tag]={'values'}
		return new_variable

	new_variable.values=np.random.standard_normal(size=structure)*random_size
	new_variable.values+=init_values
	if dropout: new_variable.tag['dropout']={'values'}
	for tag in tags: new_variable.tag['variable']={'values'}
	
	if fix: new_variable.fix()
	return new_variable

def np_to_variable_layer(variable_np, dropout=False, fix=False, tags={'variable'}):
	new_variable = variable_layer()
	new_variable.values=variable_np
	if dropout: new_variable.tag['dropout']={'values'}
	for tag in tags: new_variable.tag[tag]={'values'}
	
	if fix: new_variable.fix()
	return new_variable

def new_nn_layer(
	structure, init_values=(0,0), random_size=(None,None), dropout=True, fix=False
):
	new_nn = nn_layer()

	if random_size[0] is None:
		new_nn.weights=np.zeros(structure)+init_values[0]
	else:
		new_nn.weights=np.random.standard_normal(size=structure)*random_size[0]
		new_nn.weights+=init_values[0]

	if random_size[1] is None:
		new_nn.bias=np.zeros((1,structure[1]))+init_values[1]
	else:
		new_nn.bias=np.random.normal(size=(1,structure[1]))*random_size[1]
		new_nn.bias+=init_values[1]

	if dropout: new_nn.tag['dropout'] = {'weights'}
	new_nn.tag['weights'] = {'weights'}
	new_nn.tag['bias'] = {'bias'}
	
	if fix: new_nn.fix()
	return new_nn

def _new_lstm_layer(
    structure,

    com_in_init=0, com_in_random=None, com_in_dropout=True, com_in_fix=False,
    com_mem_init=0, com_mem_random=None, com_mem_dropout=True, com_mem_fix=False,
    com_out_init=0, com_out_random=None, com_out_dropout=True, com_out_fix=False,
    com_bias_init=0, com_bias_random=None, com_bias_dropout=False, com_bias_fix=False,

    ingate_in_init=0, ingate_in_random=None, ingate_in_dropout=True, ingate_in_fix=False,
    ingate_mem_init=0, ingate_mem_random=None, ingate_mem_dropout=True, ingate_mem_fix=False,
    ingate_out_init=0, ingate_out_random=None, ingate_out_dropout=True, ingate_out_fix=False,
    ingate_bias_init=0, ingate_bias_random=None, ingate_bias_dropout=False, ingate_bias_fix=False,

    outgate_in_init=0, outgate_in_random=None, outgate_in_dropout=True, outgate_in_fix=False,
    outgate_mem_init=0, outgate_mem_random=None, outgate_mem_dropout=True, outgate_mem_fix=False,
    outgate_out_init=0, outgate_out_random=None, outgate_out_dropout=True, outgate_out_fix=False,
    outgate_bias_init=0, outgate_bias_random=None, outgate_bias_dropout=False, outgate_bias_fix=False,

    forgate_in_init=0, forgate_in_random=None, forgate_in_dropout=True, forgate_in_fix=False,
    forgate_mem_init=0, forgate_mem_random=None, forgate_mem_dropout=True, forgate_mem_fix=False,
    forgate_out_init=0, forgate_out_random=None, forgate_out_dropout=True, forgate_out_fix=False,
    forgate_bias_init=0, forgate_bias_random=None, forgate_bias_dropout=False, forgate_bias_fix=False,

    init_mem_init=0, init_mem_random=None, init_mem_dropout=False, init_mem_fix=False,
    init_out_init=0, init_out_random=None, init_out_dropout=False, init_out_fix=False,

    com_func=uf._tanh,
    ingate_func=uf._sigmoid,
    outgate_func=uf._sigmoid,
    forgate_func=uf._sigmoid,
    final_func=uf._tanh
):

    return_lstm=lstm_layer()
    (input_size, output_size)=structure

    return_lstm.com_in=new_variable_layer(
        (input_size,output_size), init_values=com_in_init, random_size=com_in_random,
        dropout=com_in_dropout, fix=com_in_fix, tags={'weights'}
    )
    return_lstm.com_mem=new_variable_layer(
        (output_size,output_size), init_values=com_mem_init, random_size=com_mem_random,
        dropout=com_mem_dropout, fix=com_mem_fix, tags={'weights'}
    )
    return_lstm.com_out=new_variable_layer(
        (output_size,output_size), init_values=com_out_init, random_size=com_out_random,
        dropout=com_out_dropout, fix=com_out_fix, tags={'weights'}
    )
    return_lstm.com_bias=new_variable_layer(
        (1,output_size), init_values=com_bias_init, random_size=com_bias_random,
        dropout=com_bias_dropout, fix=com_bias_fix, tags={'bias'}
    )

    return_lstm.ingate_in=new_variable_layer(
        (input_size,output_size), init_values=ingate_in_init, random_size=ingate_in_random,
        dropout=ingate_in_dropout, fix=ingate_in_fix, tags={'weights'}
    )
    return_lstm.ingate_mem=new_variable_layer(
        (output_size,output_size), init_values=ingate_mem_init, random_size=ingate_mem_random,
        dropout=ingate_mem_dropout, fix=ingate_mem_fix, tags={'weights'}
    )
    return_lstm.ingate_out=new_variable_layer(
        (output_size,output_size), init_values=ingate_out_init, random_size=ingate_out_random,
        dropout=ingate_out_dropout, fix=ingate_out_fix, tags={'weights'}
    )
    return_lstm.ingate_bias=new_variable_layer(
        (1,output_size), init_values=ingate_bias_init, random_size=ingate_bias_random,
        dropout=ingate_bias_dropout, fix=ingate_bias_fix, tags={'bias'}
    )

    return_lstm.outgate_in=new_variable_layer(
        (input_size,output_size), init_values=outgate_in_init, random_size=outgate_in_random,
        dropout=outgate_in_dropout, fix=outgate_in_fix, tags={'weights'}
    )
    return_lstm.outgate_mem=new_variable_layer(
        (output_size,output_size), init_values=outgate_mem_init, random_size=outgate_mem_random,
        dropout=outgate_mem_dropout, fix=outgate_mem_fix, tags={'weights'}
    )
    return_lstm.outgate_out=new_variable_layer(
        (output_size,output_size), init_values=outgate_out_init, random_size=outgate_out_random,
        dropout=outgate_out_dropout, fix=outgate_out_fix, tags={'weights'}
    )
    return_lstm.outgate_bias=new_variable_layer(
        (1,output_size), init_values=outgate_bias_init, random_size=outgate_bias_random,
        dropout=outgate_bias_dropout, fix=outgate_bias_fix, tags={'bias'}
    )

    return_lstm.forgate_in=new_variable_layer(
        (input_size,output_size), init_values=forgate_in_init, random_size=forgate_in_random,
        dropout=forgate_in_dropout, fix=forgate_in_fix, tags={'weights'}
    )
    return_lstm.forgate_mem=new_variable_layer(
        (output_size,output_size), init_values=forgate_mem_init, random_size=forgate_mem_random,
        dropout=forgate_mem_dropout, fix=forgate_mem_fix, tags={'weights'}
    )
    return_lstm.forgate_out=new_variable_layer(
        (output_size,output_size), init_values=forgate_out_init, random_size=forgate_out_random,
        dropout=forgate_out_dropout, fix=forgate_out_fix, tags={'weights'}
    )
    return_lstm.forgate_bias=new_variable_layer(
        (1,output_size), init_values=forgate_bias_init, random_size=forgate_bias_random,
        dropout=forgate_bias_dropout, fix=forgate_bias_fix, tags={'bias'}
    )

    return_lstm.init_mem=new_variable_layer(
        (1,output_size), init_values=init_mem_init, random_size=init_mem_random,
        dropout=init_mem_dropout, fix=init_mem_fix, tags={'init_state'}
    )
    return_lstm.init_out=new_variable_layer(
        (1,output_size), init_values=init_out_init, random_size=init_out_random,
        dropout=init_out_dropout, fix=init_out_fix, tags={'init_state'}
    )

    return_lstm.com_func=com_func
    return_lstm.ingate_func=ingate_func
    return_lstm.outgate_func=outgate_func
    return_lstm.forgate_func=forgate_func
    return_lstm.final_func=final_func

    return return_lstm

def new_lstm_layer(
    structure,
    compute_init=(0,0), compute_random=(None,None),
    input_init=(0,0), input_random=(None,None),
    output_init=(0,0), output_random=(None,None),
    forget_init=(0,0), forget_random=(None,None),
    state_init=0, state_random=None,
    input_data_dropout=True,
    mem_data_dropout=False,
    reinput_data_dropout=False,
    com_func=uf._tanh,
    ingate_func=uf._sigmoid,
    outgate_func=uf._sigmoid,
    forgate_func=uf._sigmoid,
    final_func=uf._tanh,
    fix=False
):

    (com_wi, com_bi)=compute_init
    (com_wr, com_br)=compute_random
    (in_wi, in_bi)=input_init
    (in_wr, in_br)=input_random
    (out_wi, out_bi)=output_init
    (out_wr, out_br)=output_random
    (for_wi, for_bi)=forget_init
    (for_wr, for_br)=forget_random

    (in_dr,mem_dr,out_dr)=(input_data_dropout,mem_data_dropout,reinput_data_dropout)

    ret_lstm=_new_lstm_layer(
        structure=structure,

        com_in_init=com_wi, com_in_random=com_wr, com_in_dropout=in_dr, com_in_fix=False,
        com_mem_init=com_wi, com_mem_random=com_wr, com_mem_dropout=mem_dr, com_mem_fix=False,
        com_out_init=com_wi, com_out_random=com_wr, com_out_dropout=out_dr, com_out_fix=False,
        com_bias_init=com_bi, com_bias_random=com_br, com_bias_dropout=False, com_bias_fix=False,

        ingate_in_init=in_wi, ingate_in_random=in_wr, ingate_in_dropout=in_dr, ingate_in_fix=False,
        ingate_mem_init=in_wi, ingate_mem_random=in_wr, ingate_mem_dropout=mem_dr, ingate_mem_fix=False,
        ingate_out_init=in_wi, ingate_out_random=in_wr, ingate_out_dropout=out_dr, ingate_out_fix=False,
        ingate_bias_init=in_bi, ingate_bias_random=in_br, ingate_bias_dropout=False, ingate_bias_fix=False,

        outgate_in_init=out_wi, outgate_in_random=out_wr, outgate_in_dropout=in_dr, outgate_in_fix=False,
        outgate_mem_init=out_wi, outgate_mem_random=out_wr, outgate_mem_dropout=mem_dr, outgate_mem_fix=False,
        outgate_out_init=out_wi, outgate_out_random=out_wr, outgate_out_dropout=out_dr, outgate_out_fix=False,
        outgate_bias_init=out_bi, outgate_bias_random=out_br, outgate_bias_dropout=False, outgate_bias_fix=False,

        forgate_in_init=for_wi, forgate_in_random=for_wr, forgate_in_dropout=in_dr, forgate_in_fix=False,
        forgate_mem_init=for_wi, forgate_mem_random=for_wr, forgate_mem_dropout=mem_dr, forgate_mem_fix=False,
        forgate_out_init=for_wi, forgate_out_random=for_wr, forgate_out_dropout=out_dr, forgate_out_fix=False,
        forgate_bias_init=for_bi, forgate_bias_random=for_br, forgate_bias_dropout=False, forgate_bias_fix=False,

        init_mem_init=state_init, init_mem_random=state_random, init_mem_dropout=False, init_mem_fix=False,
        init_out_init=state_init, init_out_random=state_random, init_out_dropout=False, init_out_fix=False,

        com_func=com_func,
        ingate_func=ingate_func,
        outgate_func=outgate_func,
        forgate_func=forgate_func,
        final_func=final_func
    )

    if fix: ret_lstm.fix()
    return ret_lstm

def new_conv_layer(
	structure, kernal_size=(3,3),
	init_values=0, random_size=None, dropout=False,
	pad_mod='valid', strides=1,
	fix=False
):
	(input_size,output_size)=structure
	filter_shape=(output_size,input_size,*kernal_size)
	
	new_conv = conv_layer()
	new_conv.conv_func=uf._convolution
	new_conv.pad_mod=pad_mod
	new_conv.strides=strides

	if random_size is None:
		new_conv.filter_np=np.zeros(filter_shape)+init_values
		if dropout: new_conv.tag['dropout']={'filter_np'}
		new_conv.tag['weights']={'filter_np'}
		new_conv.tag['convolution']={'filter_np'}
		
		if fix: new_conv.fix()
		return new_conv

	new_conv.filter_np=np.random.standard_normal(size=filter_shape)*random_size
	new_conv.filter_np+=init_values
	if dropout: new_conv.tag['dropout']={'filter_np'}
	new_conv.tag['weights']={'filter_np'}
	new_conv.tag['convolution']={'filter_np'}
	
	if fix: new_conv.fix()
	return new_conv
	



from copy import deepcopy
import types


class folder:
	def __init__(self):
		self.mod_dict={}
		self.group_dict={}
		self.var_dict={}
		
		self.mod_dict['_is_fix']=False

	def __repr__(self):
		return f"folder({self.mod_dict})"

	@property
	def is_fix(self):
		return self.mod_dict['_is_fix']

	@is_fix.setter
	def is_fix(self, status):
		self.mod_dict['_is_fix']=status

	def fix(self):
		self.is_fix=True

	def unfix(self):
		self.is_fix=False
		self.mod_dict=deepcopy(self.mod_dict)

	@property
	def mod_name_set(self):
		return set(self.mod_dict.keys())

	@property
	def mod_with_tag_set(self):
		return self.search_mod_name_set(self.tag_name_set)

	@property
	def mod_without_tag_set(self):
		return self.mod_name_set-self.mod_with_tag_set

	@property
	def tag_name_set(self):
		return set(self.group_dict.keys())

	@property
	def tag(self):
		return tag_interface(self)

	@property
	def all_tag(self):
		return self.tag[self.tag_name_set]

	@property
	def mod(self):
		return mod_interface(self)

	def search_mod_name_set(self, tag_set):
		mod_name_set=set()
		tag_name_set=self.tag_name_set
		for tag_name in tag_set:
			if not tag_name in tag_name_set: continue
			mod_name_set|=self.group_dict[tag_name]
		return mod_name_set

	def remove_mod(self, mod_name):
		if mod_name not in self.mod_dict: return

		del self.mod_dict[mod_name]
		for tag_name in list(self.tag_name_set):
			if self.group_dict[tag_name] == {mod_name}:
				del self.group_dict[tag_name]; continue
			self.group_dict[tag_name].discard(mod_name)

	def copy(self):
		return self.all_tag.apply(copy_func, copy_func)

	def copy_without_mod(self):
		ret=type(self)()
		ret.var_dict=self.var_dict
		ret.group_dict=self.group_dict
		return ret


class mod_interface:
	def __init__(self, folder_ref):
		self.folder_ref=folder_ref

	def __getitem__(self, mod_name):
		return self.folder_ref.mod_dict[mod_name]

	def __setitem__(self, mod_name, mod):
		self.folder_ref.remove_mod(mod_name)
		self.folder_ref.mod_dict[mod_name]=mod

		if not isinstance(mod, folder): return

		for tag_name in mod.tag_name_set:
			if tag_name in self.folder_ref.tag_name_set:
				self.folder_ref.group_dict[tag_name].add(mod_name)
			else:
				self.folder_ref.tag[tag_name]={mod_name}


class tag_interface:
	def __init__(self, folder_ref):
		self.folder_ref = folder_ref

	def __repr__(self):
		return f"tag(tag_set={self.folder_ref.tag_name_set})"

	def __getitem__(self, tag_set):
		if type(tag_set)==str: tag_set={tag_set}
		return tag_operate(self.folder_ref, tag_set)

	def __setitem__(self, tag_name, group_set):
		if len(group_set-self.folder_ref.mod_name_set)>0:
			raise ValueError('group_set have undef mod!')
		self.folder_ref.group_dict[tag_name]=group_set


def copy_func(mod):
	return deepcopy(mod)

def zero_func(mod):
	if type(mod)==types.MethodType: return mod
	if type(mod)==types.FunctionType: return mod
	if type(mod)==types.MethodWrapperType: return mod
	return mod*0


class tag_operate:
	def __init__(self, folder_ref, tag_set):
		self.folder_ref=folder_ref
		self.tag_set=tag_set

	def apply(self, inset_func, outset_func=copy_func):        
		return apply_operate(
			ref=self.folder_ref,
			tag_set=self.tag_set,
			inf=inset_func,
			outf=outset_func
		)

	def dual(self, other, inset_dual_func, outset_func=copy_func):       
		return dual_operate(
			left=self.folder_ref,
			right=other,
			tag_set=self.tag_set,
			dual_inf=inset_dual_func,
			outf=outset_func
		)


def not_tag_apply_operate(ref, func):
	new_folder=ref.copy_without_mod()

	if ref.is_fix:
		new_folder.mod_dict=ref.mod_dict
		return new_folder

	mod_without_tag_set=ref.mod_without_tag_set
	for mod_name, mod in ref.mod_dict.items():
		if mod_name in mod_without_tag_set:
			new_folder.mod_dict[mod_name]=deepcopy(mod)
			continue

		if isinstance(mod, folder):
			sub_folder=not_tag_apply_operate(mod, func)
			new_folder.mod_dict[mod_name]=sub_folder
			continue

		new_folder.mod_dict[mod_name]=func(mod)
	return new_folder

def apply_operate(ref, tag_set, inf, outf):
	inside_name_set=ref.search_mod_name_set(tag_set)
	new_folder=ref.copy_without_mod()

	if ref.is_fix:
		new_folder.mod_dict=ref.mod_dict
		return new_folder

	mod_without_tag_set=ref.mod_without_tag_set
	for mod_name, mod in ref.mod_dict.items():
		if mod_name in mod_without_tag_set:
			new_folder.mod_dict[mod_name]=deepcopy(mod)
			continue

		mod_is_inside=mod_name in inside_name_set
		mod_is_folder=isinstance(mod, folder)

		if mod_is_inside and not mod_is_folder:
			new_folder.mod_dict[mod_name]=inf(mod)

		elif not mod_is_inside and not mod_is_folder:
			new_folder.mod_dict[mod_name]=outf(mod)

		elif mod_is_inside and mod_is_folder:
			sub_folder=apply_operate(mod, tag_set, inf, outf)
			new_folder.mod_dict[mod_name]=sub_folder

		else: # not mod_is_inside and mod_is_folder
			sub_folder=not_tag_apply_operate(mod, outf)
			new_folder.mod_dict[mod_name]=sub_folder

	return new_folder


def dual_operate(left, right, tag_set, dual_inf, outf):

	if not isinstance(right, folder):
		new_dual_inf=lambda mod: dual_inf(mod, right)
		return apply_operate(left, tag_set, new_dual_inf, outf)

	inside_name_set=left.search_mod_name_set(tag_set)
	new_folder=left.copy_without_mod()

	if left.is_fix:
		new_folder.mod_dict=left.mod_dict
		return new_folder

	mod_without_tag_set=left.mod_without_tag_set
	for mod_name, left_mod in left.mod_dict.items():
		if mod_name in mod_without_tag_set:
			new_folder.mod_dict[mod_name]=deepcopy(left_mod)
			continue

		right_mod=right.mod[mod_name]
		mod_is_inside=mod_name in inside_name_set
		left_mod_is_folder=isinstance(left_mod, folder)

		if mod_is_inside and not left_mod_is_folder:
			new_folder.mod[mod_name]=dual_inf(left_mod,right_mod)

		elif not mod_is_inside and not left_mod_is_folder:
			new_folder.mod[mod_name]=outf(left_mod)

		elif mod_is_inside and left_mod_is_folder:
			sub_folder=dual_operate(left_mod,right_mod,tag_set,dual_inf,outf)
			new_folder.mod[mod_name]=sub_folder

		else:
			sub_folder=not_tag_apply_operate(left_mod, outf)
			new_folder.mod[mod_name]=sub_folder

	return new_folder



from nyto import net
from nyto import folder
import random
import numpy as np

new_net=net.new_net
create_connecter=net.create_connecter

def net_get(node_if):
	node_id=node_if.node_id	
	return net.get_and_update_temp(
		net_ref=node_if.net_ref,
		node_id=node_id,
		get_temp_dict={}
	)

def net_gets(*node_ifs):
	id_dict={} # {net_id: {node_id1, node_id2}, ...}
	ref_dict={} # {net_id: net_ref}
	for node_if in node_ifs:
		net_id=id(node_if.net_ref)
		node_id=node_if.node_id

		if not net_id in id_dict:
			id_dict[net_id]=set()
			ref_dict[net_id]=node_if.net_ref

		id_dict[net_id]|={node_id}

	ret_dict={} # {net_id: {node_id: node_ret } }
	for net_id,node_id_set in id_dict.items():
		# ret={node_id: node_ret }
		ret=ref_dict[net_id].get_batch(node_id_set)
		ret_dict[net_id]=ret

	return tuple(
		ret_dict[id(node_if.net_ref)][node_if.node_id]
		for node_if in node_ifs
	)

def get(*node_ifs):
	if len(node_ifs)==1:
		return net_get(node_ifs[0])
	return net_gets(*node_ifs)


class mod_with_tag(net.add_unit_to_net):
	def __init__(self, mod, tag_set={'def'}):
		self.mod=mod
		self.tag_set=tag_set

	def __call__(self, net_ref, node_id):
		net_ref.mod[node_id]=self.mod

		for add_tag in self.tag_set:
			if not add_tag in net_ref.tag_name_set:
				net_ref.tag[add_tag]={node_id}; continue
			net_ref.group_dict[add_tag]|={node_id}

		net_ref.node[node_id]=net.mod_node(node_id)

def add_mod(mod, tag_set):
	return mod_with_tag(mod=mod, tag_set=tag_set)

def add_data(data):
	return net.static_unit(data=data)

def _get_data_size(push_dict):
	data=list(push_dict.values())[0]
	if isinstance(data, net.node_interface):
		return len(get(data))
	return len(data)

def batch_launcher(nn, batch_push, get, batch_size, static_push={}, shuffle=True):

	data_size=_get_data_size(batch_push)
	idx_list=list(range(data_size))

	if shuffle:
		if data_size%batch_size!=0:
			excess_n=batch_size-data_size%batch_size
			idx_list+=random.sample(idx_list, excess_n)

		random.shuffle(idx_list)

	batch_n=int(len(idx_list)/batch_size)
	idx_np=np.array(idx_list).reshape(batch_n, batch_size)

	launcher_node_list=[]
	for idx in idx_np:
		push={
			k:v[idx] for k,v in batch_push.items()
		}
		push={**push, **static_push}

		batch_node=nn.launcher(*get, **push)
		launcher_node_list.append(batch_node)

	return launcher_node_list

def fix_mod(net_ref, node_id):
	net_ref.mod[node_id].fix()

def unfix_mod(net_ref, node_id):
	net_ref.mod[node_id].unfix()

def free_unused_nodes(net_ref, user_node_id_set):
	used_node_id_set=net_ref.unit.name_set

	def visit_node(node_id_set):
		nonlocal used_node_id_set
		used_node_id_set|=node_id_set

		for node_id in node_id_set:
			next_node_id_set=net_ref.node[node_id].connect_node_id_set
			if len(next_node_id_set)==0: continue
			visit_node(next_node_id_set)

	visit_node(user_node_id_set)
	unused_node_id_set=net_ref.node_id_set-(net_ref.node_id_set&used_node_id_set)
	for unused_node_id in unused_node_id_set:
		del net_ref.node[unused_node_id]
	return unused_node_id_set



from nyto import folder
from nyto import particle
from nyto import net
from nyto import unit_function
from nyto import layer
from nyto import net_tool
from nyto import pool_tool
from nyto import train_tool




from nyto import folder
from copy import deepcopy


class particle(folder.folder):
    def __init__(self):
        super().__init__()
        
    def apply(self, inset_func, outset_func=folder.copy_func):
        return self.all_tag.apply(
            inset_func=inset_func,
            outset_func=outset_func
        )
    
    def dual(self, other, inset_dual_func, outset_func=folder.copy_func):
        return self.all_tag.dual(
            other=other,
            inset_dual_func=inset_dual_func,
            outset_func=outset_func
        )
        
    def __neg__(self):
        return self.apply(lambda x:-x)
    
    def __pow__(self, that):
        return self.apply(lambda x:x**that)
    
    def __add__(self, that):
        return self.dual(that, lambda s,t:s+t)
	
    def __radd__(self, that):
        return self+that
    
    def __sub__(self, that):
        return self.dual(that, lambda s,t:s-t)
	
    def __rsub__(self, that):
        return -self+that
    
    def __mul__(self, that):
        return self.dual(that, lambda s,t:s*t)
	
    def __rmul__(self, that):
        return self*that
    
    def __truediv__(self, that):
        return self.dual(that, lambda s,t:s/t)
	
    def __rtruediv__(self, that):
        return that * (self**-1)   




from nyto import net_tool
from nyto import pool_tool
import numpy as np
import random

def _net_random_normal(net_ref, tag_set=None):
	def random_mod(mod):
		if type(mod)==np.ndarray:
			return np.random.standard_normal(size=mod.shape)
		return np.random.standard_normal()

	if tag_set is None:
		return net_ref.apply(random_mod)
	return net_ref.tag[tag_set].apply(random_mod)

def _net_random(net_ref, tag_set=None):

	def random_mod(mod):
		if type(mod)==np.ndarray:
			return np.random.random(size=mod.shape)
		return np.random.random()

	if tag_set is None:
		return net_ref.apply(random_mod)
	return net_ref.tag[tag_set].apply(random_mod)

def _net_norm(net_ref, tag_set=None):

	total=0
	def sum_func(mod):
		nonlocal total
		if type(mod)==np.ndarray:
			total+=np.sum(np.square(mod)); return
		total+=mod**2

	if tag_set is None:
		net_ref.apply(sum_func)
		return np.sqrt(total)

	net_ref.tag[tag_set].apply(sum_func)
	return np.sqrt(total)

def _net_unitization(net_ref, tag_set=None):
	norm=_net_norm(net_ref=net_ref, tag_set=tag_set)

	def to_zero(mod):
		if type(mod)==np.ndarray:
			return np.zeros(shape=mod.shape)
		return 0

	def unitization(mod):
		return mod/norm

	if tag_set is None and norm==0:
		return net_ref(to_zero,to_zero)
	if tag_set is None:
		return net_ref.apply(unitization,to_zero)

	if norm==0:
		return net_ref.tag[tag_set].apply(to_zero,to_zero)
	return net_ref.tag[tag_set].apply(unitization,to_zero)

def _net_dropout_bool(net_ref, dropout_rate, dropout_tag_set={'dropout'}):

	def to_one(mod):
		if type(mod)==np.ndarray:
			return np.ones(shape=mod.shape)
		return 1

	def dropout_to_bool(mod):
		if type(mod)==np.ndarray:
			return np.random.random(size=mod.shape)>dropout_rate
		return np.random.random()>dropout_rate

	return net_ref.tag[dropout_tag_set].apply(
		inset_func=dropout_to_bool,
		outset_func=to_one
	)

def random_particle(particle_ref):
	return _net_random(particle_ref)

def normal_particle(particle_ref):
	return _net_random_normal(particle_ref)

def unit_particle(particle_ref):
	random_net=random_particle(particle_ref)
	return _net_unitization(random_net)

def unit_vector(net_ref, tag_set=None):
	random_net=_net_random_normal(net_ref, tag_set)
	return _net_unitization(random_net, tag_set)

def unit_dropout_vector(net_ref, dropout_rate=0.5, dropout_tag_set={'dropout'}):
	bool_net=_net_dropout_bool(net_ref, dropout_rate, dropout_tag_set)
	random_net=_net_random_normal(net_ref)

	dropout_vector=_net_unitization(random_net*bool_net)
	return bool_net, dropout_vector

def step_nn(nn0, unit_nn, node_id, f0=None, delta=0.01, step_rate=1, bound=None):

	delta_nn=unit_nn.all_tag.apply(lambda x: delta*x)
	nn1=nn0.all_tag.dual(delta_nn, lambda x,y: x+y)
	nn2=nn1.all_tag.dual(delta_nn, lambda x,y: x+y)

	f0=net_tool.net_get(nn0[node_id]) if f0 is None else f0
	f1=net_tool.net_get(nn1[node_id])
	f2=net_tool.net_get(nn2[node_id])

	cost1=f1-f0
	cost2=abs(f2-2*f1+f0)

	if cost2==0: return delta_nn*0

	step=(cost1/cost2)*step_rate

	if not bound is None: step=np.clip(step, a_min=-bound, a_max=bound)

	return step*delta_nn

def dropout_step_nn(nn0, random_nn, bool_nn, node_id, delta=0.01, step_rate=1, bound=None):

	drop_nn0=nn0*bool_nn
	drop_unit_nn=_net_unitization(random_nn*bool_nn)

	return step_nn(
		nn0=drop_nn0,
		unit_nn=drop_unit_nn,
		node_id=node_id,
		delta=delta, step_rate=step_rate, bound=bound
	)

def inverse_dropout(nn, dropout_rate):
	return nn.tag[{'dropout'}].apply(lambda x:x*(1-dropout_rate))

def new_pool(node_if, pool_size, random_size=1, keep=False):
	nn=node_if.net_ref
	node_id=node_if.node_id

	if keep:
		nn_list=[
			nn+random_size*_net_random_normal(nn)
			for i in range(pool_size-1)
		]+[nn.copy()]
	else:
		nn_list=[
			nn+random_size*_net_random_normal(nn)
			for i in range(pool_size)
		] 

	return pool_tool.create_pool(nn_list=nn_list, node_id=node_id)

def train_batch(batch_node_list, pool, opt, step=1, epoch=1):
	new_pool=pool.copy()
	for ep in range(epoch):
		for idx, node_if in enumerate(batch_node_list):
			node_id=node_if.node_id
			new_pool=new_pool.reset_node_id(node_id)

			new_pool=opt(new_pool)
			if (idx+1)%step==0: yield (ep, idx), new_pool, opt

	yield (ep, idx), new_pool, opt

class pool_optimization:
	pass

class epso_opt(pool_optimization):
	def __init__(self, threshold=1, dropout_rate=None, step_rate=1, bound=100):
		self.threshold=threshold
		self.dropout_rate=dropout_rate
		self.step_rate=step_rate
		self.bound=bound
		self.delta=0.01
		self.half=False
		self.adjustment=True

	def __call__(self, pool):
		node_id=pool.node_id

		if self.dropout_rate is None:
			seed_size=np.sum(pool.loss.normal<self.threshold)
			if seed_size==0: seed_size=len(pool)
			seed_list=pool.net[:seed_size]

			random_mod=lambda mod,that: (mod+that)/2
			refresh_list=[
				nn.dual(random.choice(seed_list), random_mod)
				for nn in pool.net[seed_size:]
			]

			new_nn_list=[
				nn-step_nn(
					nn0=nn,
					unit_nn=unit_vector(nn),
					node_id=node_id,
					delta=self.delta,
					step_rate=self.step_rate,
					bound=self.bound
				)
				for nn in seed_list+refresh_list
			]
			return pool_tool.create_pool(
				nn_list=new_nn_list,
				node_id=node_id
			)

		if self.half:
			seed_size=np.sum(pool.loss.normal<self.threshold)
			if seed_size==0: seed_size=len(pool)
			seed_list=pool.net[:seed_size]

			random_mod=lambda mod,that: (mod+that)/2
			refresh_list=[
				nn.dual(random.choice(seed_list), random_mod)
				for nn in pool.net[seed_size:]
			]
			
			bool_nn=_net_dropout_bool(
				net_ref=pool.net[0],
				dropout_rate=self.dropout_rate
			)

			new_nn_list=[
				nn-dropout_step_nn(
					nn0=nn,
					random_nn=_net_random_normal(nn),
					bool_nn=bool_nn,
					node_id=node_id,
					delta=self.delta,
					step_rate=self.step_rate,
					bound=self.bound
				)
				for nn in seed_list+refresh_list
			]

			return pool_tool.create_pool(
				nn_list=new_nn_list,
				node_id=node_id
			)
		
		bool_nn=_net_dropout_bool(
			net_ref=pool.net[0],
			dropout_rate=self.dropout_rate
		)
		
		if self.adjustment:
			drop_nn_pool=pool_tool.create_pool(
				nn_list=[
					nn.tag[{'dropout'}].apply(
						lambda x: x*(1-self.dropout_rate)
					)
					for nn in pool.net.real
				],
				node_id=node_id
			)
		else:
			drop_nn_pool=pool_tool.create_pool(
				nn_list=[nn*bool_nn for nn in pool.net.real],
				node_id=node_id
			)
		
		rank_pool=pool_tool.create_pool(
			nn_list=pool.net.real,
			loss_list=drop_nn_pool.loss.real,
			node_id=node_id
		)

		seed_size=np.sum(rank_pool.loss.normal<self.threshold)
		if seed_size==0: seed_size=len(pool)
		seed_list=rank_pool.net[:seed_size]

		random_mod=lambda mod,that: (mod+that)/2
		refresh_list=[
			nn.dual(random.choice(seed_list), random_mod)
			for nn in rank_pool.net[seed_size:]
		]

		new_nn_list=[
			nn-dropout_step_nn(
				nn0=nn,
				random_nn=_net_random_normal(nn),
				bool_nn=bool_nn,
				node_id=node_id,
				delta=self.delta,
				step_rate=self.step_rate,
				bound=self.bound
			)
			for nn in seed_list+refresh_list
		]

		return pool_tool.create_pool(
			nn_list=new_nn_list,
			node_id=node_id
		)
	




from nyto import particle
import types


def get_and_update_temp(net_ref, node_id, get_temp_dict):
	if node_id in get_temp_dict: return get_temp_dict[node_id]

	get_temp_dict[node_id]=net_ref.node[node_id].get(net_ref, get_temp_dict)

	return get_temp_dict[node_id]


class net_node:
	def __init__(self, node_id):
		self.node_id=node_id

class unit_node(net_node):
	@property
	def connect_node_id_set(self):
		return set()

class mod_node(unit_node):
	def get(self, net_ref, get_temp_dict):
		return net_ref.mod[self.node_id]

class data_node(unit_node):
	def get(self, net_ref, get_temp_dict):
		return net_ref.data[self.node_id]

class connect_node(net_node):
	def __init__(self, node_id, connected_node_id):
		super().__init__(node_id)
		self.connected_node_id=connected_node_id

	def get(self, net_ref, get_temp_dict): 
		return get_and_update_temp(
			net_ref,
			self.connected_node_id,
			get_temp_dict
		)

	@property
	def connect_node_id_set(self):
		return {self.connected_node_id}

class method_node(net_node):
	def __init__(self, node_id, mod_name, method_str, args, kwargs):
		super().__init__(node_id)
		self.mod_name=mod_name
		self.method_str=method_str
		self.input_args=[
			net_node(data.node_id)
			if type(data)==node_interface
			else data 
			for data in args
		]
		self.input_kwargs=dict(
			(key, net_node(value.node_id))
			if type(value)==node_interface
			else (key, value)
			for key, value in kwargs.items()
		)

	def get(self, net_ref, get_temp_dict):
		mod=get_and_update_temp(net_ref, self.mod_name, get_temp_dict)
		new_args=[
			get_and_update_temp(net_ref, data.node_id, get_temp_dict)
			if isinstance(data, net_node)
			else data
			for data in self.input_args
		]
		new_kwargs=dict(
			(
				key, 
				get_and_update_temp(net_ref, value.node_id, get_temp_dict)
			)
			if isinstance(value, net_node)
			else (key, value)
			for key, value in self.input_kwargs.items()
		)

		return_method=getattr(mod, self.method_str)
		
		if callable(return_method):
			return return_method(*new_args, **new_kwargs)
		return return_method

	@property
	def connect_node_id_set(self): 
		input_args_node_id_set={
			data.node_id 
			for data in self.input_args
			if isinstance(data, net_node)
		}
		input_kwargs_node_id_set={
			value.node_id
			for key, value in self.input_kwargs.items()
			if isinstance(value, net_node)
		}
		return set.union(
			input_args_node_id_set,
			input_kwargs_node_id_set,
			{self.mod_name}
		)

class launcher_node(net_node):
	def __init__(self, node_id, get_node_id_set, push_dict):
		super().__init__(node_id)
		self.get_node_id_set=get_node_id_set
		self.push_dict=push_dict

	def get(self, net_ref, get_temp_dict):
		net_ref.push_batch(**dict(
			(push_node_id, net_ref[push_obj.node_id])
			if isinstance(push_obj, net_node)
			else (push_node_id, push_obj)
			for push_node_id, push_obj in self.push_dict.items()
		))
		ret_dict=net_ref.get_batch(self.get_node_id_set)
		return ret_dict

	@property
	def connect_node_id_set(self):
		return set.union(
			self.pre_get_node_id_set,
			self.get_node_id_set,
			set(self.push_dict.keys())
		)

	@property
	def pre_get_node_id_set(self):
		return {
			push_obj.node_id
			for push_node_id, push_obj in self.push_dict.items()
			if isinstance(push_obj, net_node)
		}

class order_node(net_node):
	def __init__(self, node_id, order_list):
		super().__init__(node_id)
		self.order_list=order_list

	def get(self, net_ref, get_temp_dict):
		return [
			net_ref.get_batch({node_id})[node_id]
			for node_id in self.order_list
		]

	@property
	def connect_node_id_set(self):
		return set(self.order_list)

class node_interface:
	def __init__(self, net_ref, node_id):
		self.net_ref = net_ref
		self.node_id = node_id

	def __getattr__(self, method_str):
		def warp(*args, **kwargs):
			for data in args:
				if type(data)!=node_interface: continue
				if not self.net_ref is data.net_ref:
					raise ValueError('net_ref should be the same!')

			for key, values in kwargs.items():
				if type(values)!=node_interface: continue
				if not self.net_ref is values.net_ref:
					raise ValueError('net_ref should be the same!')

			new_id=self.net_ref.new_id()
			mod_name=self.node_id

			new_method_node=method_node(
				new_id, mod_name,
				method_str, args, kwargs
			)

			self.net_ref.node[new_id]=new_method_node
			return node_interface(self.net_ref, new_id)

		return warp

	def __rshift__(self, input_node_if):

		if isinstance(input_node_if, add_func_to_net):
			return input_node_if(self)

		if not input_node_if.node_id in self.net_ref.unit.name_set:
			self.net_ref[input_node_if.node_id]=self
			return self.net_ref[input_node_if.node_id]

		return input_node_if(self)

	def __call__(self, *args, **kwargs):
		return self.__getattr__('__call__')(*args, **kwargs)

	def __getitem__(self, key):
		return self.__getattr__('__getitem__')(key)

	def __neg__(self):
		return self.__getattr__('__neg__')()

	def __pow__(self, that):
		return self.__getattr__('__pow__')(that)

	def __add__(self, that):
		return self.__getattr__('__add__')(that)
		
	def __radd__(self, that):
		return self.__getattr__('__radd__')(that)

	def __sub__(self, that):
		return self.__getattr__('__sub__')(that)
		
	def __rsub__(self, that):
		return self.__getattr__('__rsub__')(that)

	def __mul__(self, that):
		return self.__getattr__('__mul__')(that)
		
	def __rmul__(self, that):
		return self.__getattr__('__rmul__')(that)

	def __truediv__(self, that):
		return self.__getattr__('__truediv__')(that)
		
	def __rtruediv__(self, that):
		return self.__getattr__('__rtruediv__')(that)

class net(particle.particle):
	def __init__(self):
		super().__init__()
		self.var_dict['node']={}
		self.var_dict['data']={}
		self.var_dict['id_count']=0

	def __repr__(self):
		return f"net(mod={self.info.mod}, data={self.info.data})"

	def __getitem__(self, node_id):
		return node_interface(self, node_id)

	def __setitem__(self, node_id, value):
		
		if type(value)==node_interface:
			if node_id in self.data_name_set:
				raise KeyError(f"data {node_id} can not overwrite!")   
			
			if node_id in self.mod_name_set:
				update=get_and_update_temp(self, value.node_id, {})
				self.unit[node_id]=update
				return
				
			self.node[node_id]=connect_node(node_id, value.node_id)
			return
		
		if node_id in self.unit:
			self.unit.remove(node_id)
			self[node_id]=value
			return

		if isinstance(value, add_unit_to_net):
			value(self, node_id)
			return
		
		self.mod[node_id]=value
		self.node[node_id]=mod_node(node_id)
	
	@property
	def info(self):
		return user_info_interface(self)

	@property
	def unit(self):    
		return unit_interface(self)

	@property
	def data(self):
		return self.var_dict['data']

	@property
	def data_name_set(self):
		return set(self.data.keys())

	@property
	def node(self):
		return self.var_dict['node']

	@property
	def node_id_set(self):
		return set(self.node.keys())

	def new_id(self):
		while True:
			return_id=self.var_dict['id_count']
			self.var_dict['id_count']+=1
			if return_id in self.node_id_set:
				continue
			return return_id
	
	def get_batch(self, node_id_set):

		return_dict={}
		get_temp_dict={}
		for node_id in node_id_set:
			return_dict[node_id]=get_and_update_temp(
				net_ref=self,
				node_id=node_id,
				get_temp_dict=get_temp_dict
			)
		return return_dict

	def push_batch(self, **push_dict):
		get_node_id_set={
			push_obj.node_id
			for push_node_id, push_obj in push_dict.items()
			if push_node_id in self.mod_name_set
			and isinstance(push_obj, node_interface)
		}
		
		get_dict=self.get_batch(get_node_id_set)
		
		new_push_dict=dict(
			(push_node_id, get_dict[push_obj.node_id])
			if push_node_id in self.mod_name_set
			and isinstance(push_obj, node_interface)
			else (push_node_id, push_obj)
			for push_node_id, push_obj in push_dict.items()
		)
		
		for push_node_id, push_obj in new_push_dict.items():
			self[push_node_id]=push_obj
	
	def push_get(self, node_id_set, **push_dict):
		self.push_batch(**push_dict)
		return self.get_batch(node_id_set)

	def launcher(self, *get_tuple, **push_dict):
		new_id=self.new_id()
		get_set={node_if.node_id for node_if in get_tuple}

		remove_node_interface_dict={}
		for push_node_id, push_obj in push_dict.items():
			if type(push_obj)==node_interface:
				push_obj=net_node(push_obj.node_id)
			remove_node_interface_dict[push_node_id]=push_obj

		self.node[new_id]=launcher_node(
			new_id, get_set, remove_node_interface_dict
		)
		return self[new_id]

	def order(self, *order_tuple):
		new_id=self.new_id()
		order_list=[node_if.node_id for node_if in order_tuple]
		self.node[new_id]=order_node(new_id, order_list)
		return self[new_id]

class unit_interface:

	def __init__(self, net_ref):
		self.ref=net_ref

	def __getitem__(self, node_id):
		if node_id in self.ref.data:
			return self.ref.data[node_id]
		return self.ref.mod[node_id]

	def __setitem__(self, node_id, value):
		if not node_id in self:
			raise KeyError(f"unit {node_id} can not find")

		if node_id in self.ref.data:
			self.ref.data[node_id]=value; return
		self.ref.mod[node_id]=value

	def __contains__(self, unit_name):
		return unit_name in self.name_set
	
	def remove(self, unit_name):
		if unit_name in self.ref.mod_name_set:
			self.ref.remove_mod(unit_name)
			del self.ref.node[unit_name]
			return
			
		if unit_name in self.ref.data_name_set:
			del self.ref.data[unit_name]
			del self.ref.node[unit_name]
			return
		
	@property
	def name_set(self):
		return self.ref.data_name_set|self.ref.mod_name_set

def _remove_hide_name(name_set):
	return {
		name
		for name in name_set
		if type(name)==str and name[0]!='_'
	}

class user_info_interface:
	def __init__(self, net_ref):
		self.ref=net_ref
	
	@property
	def unit(self):
		return _remove_hide_name(self.ref.unit.name_set)
	
	@property
	def mod(self):
		return _remove_hide_name(self.ref.mod_name_set)
	
	@property
	def data(self):
		return _remove_hide_name(self.ref.data_name_set)
	
	@property
	def node(self):
		return _remove_hide_name(self.ref.node_id_set)
	
	def is_unit(self, unit_name):
		return unit_name in self.unit
	
	def is_mod(self, mod_name):
		return mod_name in self.mod
	
	def is_data(self, data_name):
		return data_name in self.data
	
	def is_connect_node(self, node_id):
		return isinstance(self.ref.node[node_id], connect_node)
	
	def connect_id_set(self, node_id):
		return _remove_hide_name(
			self.ref.node[node_id].connect_node_id_set
		)
	
	def mod_is_fix(self, mod_name):
		return self.ref.mod[mod_name].is_fix

class add_unit_to_net:
	def __call__(self, net_ref, node_id):
		pass

class static_unit(add_unit_to_net):
	def __init__(self, data):
		self.data=data

	def __call__(self, net_ref, node_id):
		net_ref.data[node_id]=self.data
		net_ref.node[node_id]=data_node(node_id)

class add_func_to_net:
	def __call__(self, node_if):
		pass

def create_connecter(net_ref):
	class node_connecter:
		def __getattr__(self, node_id):
			return net_ref[node_id]
		def __setattr__(self, node_id, that):
			net_ref[node_id]=that
	return node_connecter()

def new_net(**push_dict):
	nn=net()
	nn.push_batch(**push_dict)
	return nn, create_connecter(nn)




from nyto import particle
from nyto import folder
from nyto import net_tool
import numpy as np


class net_pool(particle.particle):
	def __init__(self):
		super().__init__()
		self.mod['loss_list']=None
		self.mod['index_list']=None
		self.mod['node_id']=None

	@property
	def loss_list(self):
		return self.mod['loss_list']

	@loss_list.setter
	def loss_list(self, x):
		self.mod['loss_list']=x

	@property
	def index_list(self):
		return self.mod['index_list']

	@index_list.setter
	def index_list(self, x):
		self.mod['index_list']=x

	@property
	def node_id(self):
		return self.mod['node_id']

	@node_id.setter
	def node_id(self, x):
		self.mod['node_id']=x

	def __len__(self):
		return len(self.nn_name_set)

	def __getitem__(self, index):
		if type(index)==int:
			raise KeyError(f"The pool key:{index} must be slice")
		nn_list=self.net[index]
		
		if self.loss_list is None:
			return create_pool(
				nn_list=nn_list,
				node_id=self.node_id
			)
		
		return create_pool(
			nn_list=nn_list,
			loss_list=self.loss[index],
			node_id=self.node_id
		)

	def __repr__(self):
		self.init_loss_index_list()
		return f"pool(loss={self.loss[:]})"

	def apply(self, inset_func, outset_func=folder.copy_func):
		ret_pool=super().apply(inset_func, outset_func)
		ret_pool.loss_list=None
		ret_pool.index_list=None
		return ret_pool

	def dual(self, other, inset_dual_func, outset_func=folder.copy_func):
		if isinstance(other, net_pool):
			return self.pool_dual(other, inset_dual_func)

		ret_pool=super().dual(other, inset_dual_func, outset_func)
		ret_pool.loss_list=None
		ret_pool.index_list=None
		return ret_pool

	def pool_apply(self, pool_func):
		ret_pool=self.copy_without_mod()
		for nn_name in self.nn_name_set:
			ret_pool.mod[nn_name]=pool_func(self.mod[nn_name])
		ret_pool.node_id=self.node_id
		return ret_pool

	def pool_dual(self, other_pool, dual_func):
		ret_pool=self.copy_without_mod()
		for nn_name in self.nn_name_set:
			self_nn=self.mod[nn_name]
			that_nn=other_pool.mod[nn_name]
			ret_pool.mod[nn_name]=dual_func(self_nn,that_nn)
		ret_pool.node_id=self.node_id
		return ret_pool

	@property
	def nn_name_set(self):
		return {
			mod_name for mod_name in self.mod_name_set
			if type(mod_name)==int
		}

	@property
	def net_list(self):
		return [
			self.mod[idx] for idx in range(len(self))
		]

	@property
	def net(self):
		self.init_loss_index_list()
		return pool_net(self)
	
	@property
	def real_net(self):
		return pool_real_net(self)
	
	@property
	def loss(self):
		self.init_loss_index_list()
		return pool_loss(self)

	def reset_node_id(self, node_id):
		return create_pool(
			nn_list=self.net_list,
			node_id=node_id
		)

	def init_loss_index_list(self):
		if self.loss_list is None:
			self.loss_list=self.get_loss_list()

		if self.index_list is None:
			self.index_list=self.get_index_list()

	def get_loss_list(self):
		loss_list=[None]*len(self)
		for nn_name in self.nn_name_set:
			loss=net_tool.net_get(self.mod[nn_name][self.node_id])
			loss_list[nn_name]=loss
		return loss_list

	def get_index_list(self):
		return sorted(
			range(len(self)),
			key=lambda k:self.loss_list[k]
		)
	

class pool_net:
	def __init__(self, pool_ref):
		self.pool_ref=pool_ref

	def __getitem__(self, idx):
		if type(idx)==slice:
			idx_list=self.pool_ref.index_list[idx]
			return [self.pool_ref.mod[idx] for idx in idx_list]
		return self.pool_ref.mod[self.pool_ref.index_list[idx]]
	
	def __setitem__(self, idx, new_net):
		if type(idx)==slice:
			raise KeyError(f'The net key:{idx} must not be slice')
		self.pool_ref.real_net[self.pool_ref.index_list[idx]]=new_net
	
	@property
	def real(self):
		return self.pool_ref.net_list

class pool_real_net:
	def __init__(self, pool_ref):
		self.pool_ref=pool_ref
		
	def __getitem__(self, idx):
		return self.pool_ref.mod[idx]
	
	def __setitem__(self, idx, new_net):
		self.pool_ref.mod[idx]=new_net
		self.pool_ref.loss_list[idx]=net_tool.net_get(
			self.pool_ref.mod[idx][self.pool_ref.node_id]
		)
		self.pool_ref.index_list=None
	
class pool_loss:
	def __init__(self, pool_ref):
		self.pool_ref=pool_ref

	def __getitem__(self, idx):
		if type(idx)==slice:
			idx_list=self.pool_ref.index_list[idx]
			return [self.pool_ref.loss_list[idx] for idx in idx_list]
		return self.pool_ref.loss_list[self.pool_ref.index_list[idx]]

	@property
	def real(self):
		return self.pool_ref.loss_list

	@property
	def mean(self):
		return np.mean(self.pool_ref.loss_list)

	@property
	def std(self):
		return np.std(self.pool_ref.loss_list)

	@property
	def normal(self):
		return (self.real-self.mean)/self.std
	

def create_pool(nn_list, node_id, loss_list=None):
	pool=net_pool()
	for idx,nn in enumerate(nn_list):
		pool.mod[idx]=nn

	pool.node_id=node_id
	pool.loss_list=loss_list
	return pool
